#!/usr/bin/python

import sys
import os
import base64
from shutil import rmtree
from subprocess import CalledProcessError

from common import write_json_file, JujuBroker, LandscapeBroker, chown
from ceph import (
    get_ceph_client_path, write_ceph_client_keyring, write_ceph_client_config)


CERT_FILE = "/etc/ssl/certs/landscape_server_ca.crt"


def _write_certificate(certificate, filename):
    """
    @param certificate Text of the certificate, base64 encoded.
    @param filename Full path to file to write
    """
    with open(filename, "w") as file:
        file.write(base64.b64decode(certificate))


def registration_relation(juju_broker, landscape_broker):
    data = juju_broker.get_relation_config()
    if "ping-url" in data:
        config_changed(juju_broker, landscape_broker, relation_data=data)


def config_changed(juju_broker, landscape_broker, relation_data=None):
    """
    @param relation_data: data from the context of a relation, it should
                          match the data in the config
    """
    if relation_data is None:
        relation_data = {}
    service_config = juju_broker.get_service_config()
    # Only update client config with service configs that have been set
    service_config = dict(
        (key, value) for key, value in service_config.iteritems()
        if value is not None)
    relation_data.update(service_config)
    if relation_data.get("ssl-public-key", "").startswith("base64:"):
        _write_certificate(relation_data["ssl-public-key"][7:], CERT_FILE)
        relation_data["ssl-public-key"] = CERT_FILE
    if "account-name" in relation_data:
        return landscape_broker.update_client_config(relation_data)


def upgrade_charm(juju_broker, landscape_broker):
    """Idempotently upgrades state from older charms."""
    landscape_broker.config.reload()
    return migrate_old_juju_info(landscape_broker.config)


def migrate_old_juju_info(client_config):
    """
    Migrates data from the old meta-data.d directory into the new
    juju-info.json file.
    """
    fallback_meta_data_dir = os.path.join(
        client_config.data_path, "meta-data.d")
    meta_data_dir = client_config.get("meta_data_path", fallback_meta_data_dir)
    if not os.path.exists(meta_data_dir):
        return 0
    juju_info_file = os.path.join(client_config.data_path, "juju-info.json")
    juju_key_map = {"juju-env-uuid": "environment-uuid",
                    "juju-unit-name": "unit-name",
                    "juju-api-addresses": "api-addresses",
                    "juju-private-address": "private-address"}
    juju_info = dict.fromkeys(juju_key_map.values())
    files_to_delete = []
    for filename, info_key in juju_key_map.iteritems():
        file_path = os.path.join(meta_data_dir, filename)
        if not os.path.exists(file_path):
            continue
        with open(file_path, "r") as juju_info_part_file:
            juju_info[info_key] = juju_info_part_file.read().strip()
        files_to_delete.append(file_path)
    missing_values = [value is None for value in juju_info.itervalues()]
    # If they're all missing it probably means we've already converted.
    if all(missing_values):
        return 0
    # ... otherwise there's something unexpected going on.
    if any(missing_values):
        return "Incomplete Juju information found in %s" % meta_data_dir
    write_json_file(juju_info_file, juju_info)
    for file_to_delete in files_to_delete:
        os.remove(file_to_delete)
    return 0


def container_relation_joined(juju_broker, landscape_broker):
    landscape_broker.config.reload()
    relation_conf = juju_broker.get_relation_config()
    # We use the remote unit for the unit name, since we want to associate this
    # client with the unit it's managing, not its own unit.
    juju_info = {
        "environment-uuid": juju_broker.environment.get("JUJU_ENV_UUID"),
        "unit-name": juju_broker.environment.get("JUJU_REMOTE_UNIT"),
        "api-addresses": juju_broker.environment.get("JUJU_API_ADDRESSES"),
        "private-address": relation_conf.get("private-address")}
    juju_info_file = os.path.join(
        landscape_broker.config.data_path, "juju-info.json")
    write_json_file(juju_info_file, juju_info)
    exit_code = landscape_broker.update_client_config(
        {"computer-title": juju_broker.environment["JUJU_REMOTE_UNIT"]})

    return exit_code


def container_relation_departed(juju_broker, landscape_broker):
    """Stop managing the removed unit with Landscape.

    If the relation is removed, the registration will be cleared, the
    computer title unset and the juju info file removed, so that we
    won't try to register again when the config-changed hook is fired.
    """
    landscape_broker.clear_registration(),
    juju_info_path = os.path.join(
        landscape_broker.config.data_path, "juju-info.json")
    if os.path.exists(juju_info_path):
        os.remove(juju_info_path)
    landscape_broker.update_client_config({"computer-title": ""})


def ceph_relation_changed(juju_broker, landscape_broker):
    relation_config = juju_broker.get_relation_config()
    # The key is only reported once the ceph ring has quorum (there are enough
    # nodes in the service). There's nothing to do until the key is provided.
    if "key" not in relation_config:
        juju_broker.log(
            "Ceph key not reported in relation config, nothing to do.")
        return 0

    ceph_client_dir = get_ceph_client_path(landscape_broker)
    if not os.path.exists(ceph_client_dir):
        os.mkdir(ceph_client_dir)
        chown(ceph_client_dir)

    # Create ceph client keyring and config for landscape-client use.
    try:
        keyring_file = write_ceph_client_keyring(
            landscape_broker, "landscape-client", relation_config["key"])
    except CalledProcessError as err:
        juju_broker.log(
            "Writing ceph keyring file failed with code %d" % err.returncode)
        return 1

    try:
        write_ceph_client_config(
            landscape_broker, relation_config["auth"], keyring_file,
            relation_config["private-address"])
    except IOError as err:
        juju_broker.log("Writing ceph configuration file failed: %s" % err)
        return 1
    else:
        juju_broker.log("Ceph configuration written to file.")
        return 0


def ceph_relation_broken(juju_broker, landscape_broker):
    rmtree(get_ceph_client_path(landscape_broker), ignore_errors=True)


class HookRunner(object):
    """A hook runner.

    The available hooks are defined in the HOOKS attribute, which is a
    dict with the hook name as the key and the hook function as the
    value. The hook function should take a juju and landscape broker as
    parameters.
    """

    HOOKS = {
        "config-changed": config_changed,
        "upgrade-charm": upgrade_charm,
        "registration-relation-joined": registration_relation,
        "registration-relation-changed": registration_relation,
        "registration-relation-departed": lambda juju_broker, landscape_broker:
            landscape_broker.clear_registration(),
        "container-relation-joined": container_relation_joined,
        "container-relation-departed": container_relation_departed,
        "ceph-client-relation-broken": ceph_relation_broken,
        "ceph-client-relation-joined": container_relation_joined,
        "ceph-client-relation-changed": ceph_relation_changed,
        "ceph-client-relation-departed": lambda juju_broker, landscape_broker:
            landscape_broker.clear_registration()}

    def __init__(self, juju_broker=None, landscape_broker=None):
        if juju_broker is None:
            juju_broker = JujuBroker()
        self.juju_broker = juju_broker
        if landscape_broker is None:
            landscape_broker = LandscapeBroker()
        self.landscape_broker = landscape_broker

    def run(self, hook_name):
        """Run the hook with the given name.

        If the hook doesn't exist, a KeyError is raised.
        """
        hook = self.HOOKS[hook_name]
        return hook(self.juju_broker, self.landscape_broker)


if __name__ == "__main__":
    hook_name = os.path.basename(sys.argv[0])
    hooks = HookRunner()
    sys.exit(hooks.run(hook_name))
