#!/bin/bash

set -eu

# Test setup
readonly TEST_NAME=$0

# Test counts. Neither skips or errors are necessarily fatal.
SKIPS=0
ERRORS=0
UNHANDLED=0


function info() {
    echo "INFO: $@"
}

function ok() {
    echo "PASS: $@"
}

function skip() {
    echo "SKIP: $@"
    (( SKIPS +=1 ))
}

function error() {
    echo "FAIL: $@"
    (( ERRORS += 1 ))
}


function start-test() {
    readonly DATADIR=$(mktemp -d --tmpdir "${TEST_NAME}.XXXXXXX")
    info "Executing ${TEST_NAME} with results in $DATADIR: $1"
    setup
    trap trapped-teardown EXIT TERM INT
}

function setup() {
    # Empty, override in actual test as needed
    true
}


function end-test() {
    info "Completed test $TEST_NAME"
    trap - EXIT
    test-teardown
}


function teardown() {
    # Empty, override in actual test as needed
    true
}


function test-teardown() {
    teardown

    if [[ "$ERRORS" -gt 0 ]] ; then
        TEST_RESULT=1
    elif [[ "$SKIPS" -gt 0 ]] ; then
        TEST_RESULT=100
        touch $DATADIR/passed
    else 
        TEST_RESULT=0
        touch $DATADIR/passed
    fi

    if [ -n "$DATADIR" ] ; then
        if [ -f $DATADIR/passed ]; then
            info "Passed test results for ${TEST_NAME} in ${DATADIR}, skips=${SKIPS}"
            # rm -r $DATADIR
        else
            info "Failed test results for ${TEST_NAME} in ${DATADIR}, errors=${ERRORS}, skips=${SKIPS}"
            if [ -f $DATADIR/wget.log ] ; then
                info "BEGIN wget.log"
                cat $DATADIR/wget.log
                info "END wget.log"
            fi
        fi
    fi
    exit $TEST_RESULT
}


function diagnose() {
    # Unpack the caller info to get some useful diagnostics for
    # tracing any assertion issues
    local data=( $(caller 1) )
    local lineno=${data[0]}
    local function_name=${data[1]}
    local script_name=${data[2]}
    echo "function $function_name at line $lineno ($script_name)"
}

function trapped-teardown() {
    (( UNHANDLED += 1 ))
    error "Untrapped error in $(diagnose)"
    test-teardown
}

function assert() {
    # The assert function takes two args, such as:
    # assert "what we are asserting" "non empty result passes"
    # If the second arg is empty or undefined, the assertion fails
    if [[ "$#" -lt 2 || -z "$2" ]] ; then
        error "$1, assertion failed in $(diagnose)"
    else
        ok "$1"
    fi
}
